/* 
 * $Id: pmap.c,v 1.5 2002/10/02 19:53:05 romanp Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/syscall.h>
#include <sys/module.h>
#include <sys/user.h>
#include <sys/vnode.h>

#include <vm/vm_object.h>
#include <vm/swap_pager.h>

#include <kvm.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "pmap_helper.h"
#include "pmap.h"

static struct kinfo_proc *kp; /* kernel process table */
u_int m_size, o_size, r_size; /* total process maps, objs, resident pages */
u_int p_size_m, p_size_r;     /* private maps, resident pages */
u_int s_size_m, s_size_r;     /* shared maps, resident pages */
u_int swap_size;              /* process swap usage */

/* total statistics for all processes */
u_int total_swap, total_rsize, total_osize, total_msize, total_psizer;

/* meanings of vm_map_entry_t->eflags */
bit_t eflags_bits[] = {	
    {"M",	0x1},
    {"SM",	0x2},
    {"COW",	0x4},
    {"NC",	0x8},
    {"NF",	0x10},
    {"UW",	0x20},
    {"", 	0},
};

/* meanings of vm_object_t->flags */
bit_t	vmobj_flags_bits[] = { 
    {"ACTIVE",0x0004},
    {"DEAD",0x0008},
    {"NOSPLIT",0x0010},
    {"PIPWNT",0x0040},
    {"WRITEABLE",0x0080},
    {"MIGHTBEDIRTY",0x0100},
    {"CLEANING",0x0200},
    {"OPT",0x1000},
    {"ONEMAPPING",0x2000},
    {"",0},
};

int 		main(int argc, char **argv);
static int 	pmap_helper(struct pmh *);
static void 	process_vm_object(struct pmh_map *);
static void 	process_swap_pager(struct vm_object *vm_o);

int 
main(int argc, char **argv)
{
    int nproc, i, prc;
    int pid = 0;

    if(argc >= 2) /* use argv[1] as pid */
	if(!(pid = atoi(argv[1]))) {
	    printf("usage: pmap [pid]\n");
	    exit(1);
	}

    /* open kernel memory */
    if(k_open() == -1) 
	exit(1);
    /* load name cache */
    namec_load();
    /* read process list */
    kp = k_getallprocs(&nproc);
    if(kp == NULL)
	exit(1);
    for(prc = 0; prc < nproc; prc++ )
    {
#define K_PPROC(a)		((a).kp_proc)
	struct pmh		pmh;
	struct pmh_map 		maps[1024];

	if(pid && (K_PPROC(kp[prc]).p_pid != pid))
	    continue;
	/* print general info */
	printf("\npid: %i comm: %s\n", 
	       K_PPROC(kp[prc]).p_pid, K_PPROC(kp[prc]).p_comm);
	/* get virtual maps for process via pmap_helper */
	pmh.version = PMH_VERSION;
	pmh.pid = K_PPROC(kp[prc]).p_pid;
	pmh.nmaps = 1024;
	pmh.maps = maps;
	if(pmap_helper(&pmh)) {
	    perror("pmap_helper");
	    continue;
	}
	printf("%8s %7s %5s %12s" 
	       "%7s %4s %5s %7s %7s %s\n",  
	       "ADDRESS","MSIZE", "MPROT", "MFLAGS", 
	       "OSIZE", "OREF", "OSREF", "ORES", "MRES", "NAME");
	printf("--------------------------------------------------------------\n");
	m_size=o_size=r_size=swap_size=0;
	p_size_r=p_size_m=s_size_m=s_size_r=0;

	for(i=0; i < pmh.nmaps; i++) {
#define MAP_VME  (maps[i].entry)
	    printf("%08x %7dK ", MAP_VME.start, 
		   (MAP_VME.end - MAP_VME.start)/1024);
	    m_size+=(MAP_VME.end-MAP_VME.start)/1024;
	    if(MAP_VME.protection & VM_PROT_READ)
		printf("r");
	    else 
		printf("-");
	    if(MAP_VME.protection & VM_PROT_WRITE)
		printf("w");
	    else
		printf("-");
	    if(MAP_VME.protection & VM_PROT_EXECUTE)
		printf("x");
	    else
		printf("-");
	    printf(" %12s ", print_bits(MAP_VME.eflags, eflags_bits));
	    process_vm_object(&maps[i]);
	    printf("\n");
	}
	printf("--------------------------------------------------------------\n");
	printf("maps size %6dK, obj size %6dK, resident pages %6dK\n", 
	       m_size, o_size, r_size);
	printf("Private pages: %dK(%dK), shared pages: %dK(%d)K, swap usage %dK\n", p_size_r, p_size_m, s_size_r, s_size_m, swap_size); 
	total_swap+=swap_size;
	total_osize+=o_size;
	total_msize+=m_size;
	total_rsize+=r_size;
	total_psizer+=p_size_r;
    }		
    k_close();
    printf("Total: %dK swap, %dK mapped memory, %dK objects memory\n"
	   "%dK resident memory (with shared), %dK procs private memory\n", 
	   total_swap, total_msize, total_osize, total_rsize, total_psizer);
    exit(0);
}

static int 
pmap_helper(struct pmh *pmh)
{
    int syscall_num;
    struct module_stat stat;

    stat.version = sizeof(stat);
    modstat(modfind("pmap_helper"), &stat);
    syscall_num = stat.data.intval;
    return syscall(syscall_num, pmh);
}

static void 
process_vm_object(struct pmh_map *map)
{
    struct vm_object *obj = &map->obj;
    struct vm_object *lobj = &map->lobj;
    struct vnode vn;
    int rc;

    printf("%7dK %3i %3i %7dK %7dK ",
	   obj->size<<2, obj->ref_count, obj->shadow_count,
	   obj->resident_page_count<<2,
	   map->resident_pages<<2);
    o_size+=(obj->size<<2);
    if(obj->shadow_count == 1) {
	r_size+=(obj->resident_page_count<<2);
	p_size_r+=(obj->resident_page_count<<2);
	p_size_m+=(map->entry.end - map->entry.start)/1024;
    } else {
	r_size+=(map->resident_pages<<2);
	s_size_r+=(map->resident_pages<<2);
	s_size_m+=(map->entry.end - map->entry.start)/1024;
    }
    switch( lobj->type ) {
    case OBJT_VNODE:
	/* if object type is OBJT_VNODE, try to read vnode */
	rc = k_read((u_long)lobj->handle, &vn, sizeof(vn));
	if(rc != -1)
	    printf("%s", namec_lookup((u_long)lobj->handle,vn.v_id));
	else
	    printf("[vnode pager, origin unknown]");
	break;
    case OBJT_SWAP:
	printf("[ swap pager ]"); 
	process_swap_pager(lobj);
	break;
    case OBJT_DEFAULT:
	printf("[ anon ]");
	break;
    case OBJT_DEVICE:
	printf("/dev/%s", map->device);
	break;
    case OBJT_DEAD:
	printf("[dead object]");
	break;
    default:
	printf("[unknown (%i) type]", lobj->type);
	break;
    }
}

static void 
process_swap_pager(struct vm_object *vm_o)
{
#ifdef notyet
    struct swblock *swblks;
    int rc, npages = 0;
    register blk, page;

    /* allocate memory for swp_blocks[] */
    swblks = malloc(vm_o.un_pager.swp.swp_nblocks*sizeof(*swblks));
    rc = kvm_read(kvmd, (u_long)vm_o.un_pager.swp.swp_blocks, swblks, 
		  vm_o.un_pager.swp.swp_nblocks*sizeof(*swblks));
#define SWB_EMPTY 	(-1)	/* from /sys/vm/swap_pager.c */
    /* count allocated pager blocks */
    for(blk=0;blk<vm_o.un_pager.swp.swp_nblocks;blk++)
	for(page=0; page < SWB_NPAGES; page ++)
	    if(swblks[blk].swb_block[page] != SWB_EMPTY)
		npages++;
    printf("%iK used]", npages*4);
    swap_size+=(npages<<2);
    free(swblks);
#endif
}
