/*-
 * Copyright (c) 2000, 2002 Roman V. Palagin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = 
  "$Id: pmap.c,v 1.9 2004/04/20 09:55:01 romanp Exp $";
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/syscall.h>
#include <sys/module.h>
#include <sys/user.h>
#include <sys/vnode.h>
#include <sys/linker.h>

#include <vm/vm_object.h>
#include <vm/swap_pager.h>

#include <kvm.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <regex.h>

#include "pmap_helper.h"
#include "pmap.h"

/* kernel process table */
static struct kinfo_proc *kp;

/* total memory usage for process */
u_int p_msize = 0, p_rsize = 0, p_ssize = 0, p_psize = 0;

/* process memory map size */
u_int	map_size;

/* pmap_helper syscall no. */
int	pmap_helper_syscall;

int 		main(int argc, char **argv);
static int 	pmap_helper(struct pmh *);
static void 	process_map(struct pmh_map *);
static void	print_proto(vm_prot_t);
static void	usage(void);

#define MAP_VME(m)  	((m).entry)
#define NKPG		(PAGE_SIZE/1024)

int 
main(int argc, char **argv)
{
	int c, nproc, i, prc;
	int pid = 0;
	struct module_stat stat;
	int modid;
	char	*prog_wild = NULL, *prog_name;
	regex_t preg;
	struct pargs *pa = NULL;

	if ((modid = modfind("pmap_helper")) == -1) { 
		/* module not found, try to load */
		modid = kldload("pmap_helper.ko");
		if (modid == -1) 
			err(1, "unable to load pmap_helper module");
		modid = modfind("pmap_helper");
		if (modid == -1)
			err(1, "pmap_helper module loaded but not found");
	}

	stat.version = sizeof(stat);
	modstat(modid, &stat);
	pmap_helper_syscall = stat.data.intval;

	while ((c = getopt(argc, argv, "n:")) != EOF) {
		switch (c) {
		case 'n':
			prog_wild = optarg;
			break;
		default:
			usage();
			break;
		}
	}
	if (prog_wild != NULL) {
		if (0 != regcomp(&preg, prog_wild,
			REG_EXTENDED | REG_ICASE | REG_NOSUB)) {
			errx(1, "regcomp() failed\n");
		}
	} else if ((argc == 2) && ((pid = atoi(argv[1])) == 0))
		usage();
		
	/* open kernel memory */
	k_open();
	/* load name cache */
	namec_load();
	/* read process list */
	kp = k_getallprocs(&nproc);

	for (prc = 0; prc < nproc; prc++) {
#define K_PPROC(a)		((a).kp_proc)
		struct pmh		pmh;
		struct pmh_map 		maps[1024];

		if ((pid != 0) && (K_PPROC(kp[prc]).p_pid != pid))
			continue;
		// try to read command line
		if (K_PPROC(kp[prc]).p_args != NULL) {
			int rc;

			pa = realloc(pa, sizeof(struct pargs));
			rc = k_read(K_PPROC(kp[prc]).p_args, pa, sizeof(struct pargs));
			if (rc != -1) {
				int news = sizeof(struct pargs) + pa->ar_length + 1;
				pa = realloc(pa, news);
				memset(pa, 0, news);
				rc = k_read(K_PPROC(kp[prc]).p_args , pa, news + 1);
			}
			prog_name = pa->ar_args;
		} else {
			prog_name = K_PPROC(kp[prc]).p_comm;

		}

		if (prog_wild != NULL) {
			if (regexec(&preg, prog_name, 0, NULL, 0) == REG_NOMATCH)
				continue;
		}

		/* print general info */
		printf("%i:\t%s\n", 
		    K_PPROC(kp[prc]).p_pid, prog_name);
		/* get virtual maps for process via pmap_helper */
		pmh.version = PMH_VERSION;
		pmh.pid = K_PPROC(kp[prc]).p_pid;
		pmh.nmaps = 1024;
		pmh.maps = maps;
		if (pmap_helper(&pmh)) {
			warn("pmap_helper");
			continue;
		}
		printf("%-8s %6s %8s %6s %7s %-15s %s\n",
		    "Address", "Kbytes", "Resident", "Shared", "Private", 
		    "Protection", "Mapped File");
		p_msize = p_rsize = p_ssize = p_psize = 0;
	
		for (i=0; i < pmh.nmaps; i++) {
			map_size = (MAP_VME(maps[i]).end - MAP_VME(maps[i]).start)/1024;
			printf("%08X %6d ", MAP_VME(maps[i]).start, map_size);
			p_msize += map_size;
			process_map(&maps[i]);
			printf("\n");
		}
		printf("%-8s %6s %8s %6s %7s\n",
		    "--------", "------", "--------", "------", "-------");
		printf("%-8s %6d %8d %6d %7d\n\n",
		    "Total", p_msize, p_rsize, p_ssize, p_psize);
	}
		
	k_close();
	if (prog_wild != NULL)
		regfree(&preg);
	exit(0);
}

static int 
pmap_helper(struct pmh *pmh)
{
	return syscall (pmap_helper_syscall, pmh);
}

static void 
process_map(struct pmh_map *map)
{
	struct vm_object *obj = &map->obj;
	struct vm_object *lobj = &map->lobj;
	struct vnode vn;
	int rc;
	struct pmh_map m = *map;

	/* sanity checks */
	if (!m.lobj_present) {
		printf("%8s %6s %7s ","-","-","-");
		print_proto(MAP_VME(m).protection);       
		printf("  [no lobj]"); 
		return;
	}

	if (!m.obj_present)
		printf("oops - no obj!");

	/* 
	 * we calculate memory usage depending on object type
	 * output: resident, shared, priv, proto, mapped file 
	 */
	switch (lobj->type) {
	case OBJT_DEFAULT: /* XXX treat all memory as private. Is it 
			      always correct? 
			   */
		printf("%8d %6s %7d ", m.resident_pages * NKPG, 
		    "-", map_size); 
		p_rsize += m.resident_pages * NKPG;
		p_psize += map_size;
		print_proto(MAP_VME(m).protection);       
		printf("  [ anon ]");
		break;

	case OBJT_VNODE:
		printf("%8d ", 
		    (obj->shadow_count == 1) ? 
		    m.resident_pages * NKPG :
		    obj->resident_page_count * NKPG);
		p_rsize +=  (obj->shadow_count == 1) ? 
		    m.resident_pages * NKPG :
			obj->resident_page_count * NKPG;
		if (obj->shadow_count == 1) {
			printf("%6s %7d ", 
			       "-", obj->resident_page_count * NKPG);
			p_psize += obj->resident_page_count * NKPG;
		} else if (obj->ref_count == 1) { /* XXX does it alawys mean private mapping (?) */
			printf("%6s %7d ", "-", map_size);
			p_psize += map_size;
		} else {
			printf("%6d %7s ", map_size, "-");
			p_ssize += map_size;
		}
		print_proto(MAP_VME(m).protection);       
		/* if object type is OBJT_VNODE, try to read vnode */
		rc = k_read((u_long)lobj->handle, &vn, sizeof(vn));
		if (rc != -1)
			printf("%s", namec_lookup((u_long)lobj->handle,vn.v_id));
		else
			printf("[vnode pager, origin unknown]");
		break;

	case OBJT_SWAP:
		printf("%8s %6s %7s ","-","-","-");
		print_proto(MAP_VME(m).protection);       
		printf("[swap pager]"); 
		break;

	case OBJT_DEVICE:
		printf("%8d %6s %7d ", m.resident_pages * NKPG, "-", map_size);
		print_proto(MAP_VME(m).protection);
		printf("/dev/%s", m.device);
		break;

	case OBJT_DEAD:
		printf("%8s %6s %7s ","-","-","-");
		print_proto(MAP_VME(m).protection);       
		printf("[dead object]");
		break;

	default:
		printf("[unknown (%i) object type]", lobj->type);
		break;
	}
}

static void
print_proto (vm_prot_t p)
{
	char str[20];

	str[0] = '\0';

	if(p & VM_PROT_READ)
		strcat(str, "read");
	if(p & VM_PROT_WRITE)
		strcat(str, "/write");
	if(p & VM_PROT_EXECUTE)
		strcat(str, "/exec");
	if (str[0] == '\0')
		sprintf(str,"0x%08x", p);
	printf("%-15s ", str);
}

static void
usage(void)
{
	errx(1, "usage: pmap [-n regexp] [pid]\n");
}
